@@ -5,6 +5,7 @@ import warnings
 from http import HTTPStatus
 from queue import LifoQueue
 from time import sleep
+from typing import Optional
 
 import numpy as np
 import requests
@@ -82,16 +83,28 @@ class Client:
             print(f"Warning: Failed to start RabbitMQ consumer: {e}")
 
         self.get_run_details()
-        return
 
-    def _get_auth_headers(self):
-        """Generate Bearer auth headers for API requests"""
+    def _get_auth_headers(self) -> dict[str, str]:
+        """
+        Generate Bearer authentication headers for API requests
+
+        Returns:
+            dict: Headers with Authorization and accept fields
+        """
         return {
             "Authorization": f"Bearer {self.__leaf_id}:{self.__leaf_session_token}",
             "accept": "application/json",
         }
 
-    def get_run_details(self):
+    def get_run_details(self) -> None:
+        """
+        Fetch current run status and run number from the API
+
+        Updates internal state with run_status and run_number
+
+        Raises:
+            Exception: If API request fails or response cannot be parsed
+        """
         try:
             url = self.__api_host + "/api/v1/run/" + self.__branch_id
             headers = self._get_auth_headers()
@@ -125,47 +138,151 @@ class Client:
         if status == "start":
             self.get_run_details()
 
-    def disable_ssl_verification(self, enabled=False):
+    def disable_ssl_verification(self, enabled: bool = False) -> bool:
+        """
+        Enable or disable SSL certificate verification
+
+        Args:
+            enabled: Whether to enable SSL verification
+
+        Returns:
+            bool: Current SSL verification state
+
+        Warning:
+            Disabling SSL verification is a security risk and not recommended
+        """
         warnings.warn(
             "[BranchKey] We highly recommend enabling ssl verification. Use this at your own risk."
         )
         self.__verify = enabled
         return self.__verify
 
-    def convert_pytorch_numpy(self, model, weighting=1):
+    def convert_pytorch_numpy(self, model, weighting: float = 1) -> tuple[float, list[np.ndarray]]:
         """
-        Usage:
-        update = convert_pytorch_numpy(model.get_params(), client_samples=5*5)
-        :param model: Pytorch model parameters
-        :param weighting: Weight given to this update. Usually the number of client samples used to construct the model
-        :return: type(np.ndarray) of parameters
+        Convert PyTorch model parameters to numpy arrays
+
+        Args:
+            model: PyTorch model named_parameters() iterator
+            weighting: Weight for aggregation (default: 1)
+                - Number of training samples (most common)
+                - Fixed value for equal weighting
+                - Quality score (e.g., validation accuracy)
+
+        Returns:
+            tuple: (weighting, list of numpy arrays)
+
+        Example:
+            >>> weighting, params = client.convert_pytorch_numpy(
+            ...     model.named_parameters(),
+            ...     weighting=len(train_dataset)
+            ... )
         """
         params = []
         for param in model:
             params.append(param[1].data.cpu().detach().numpy())
-        numpy = (weighting, params)
-        return np.asarray(numpy, dtype=object)
+        return (weighting, params)
 
-    def validate_weight_shape(self, file):
+    def save_weights(self, file_path: str, weighting: float, parameters: list[np.ndarray]) -> str:
         """
-        Validates that the weights are in the correct shape before being sent to the API
-        :param file: file bytes or file path
-        :return: True on success, otherwise returns an exception
-        """
-        array = np.load(file, allow_pickle=True)
-        if array.shape[0] != 2:
-            raise Exception("[validate_weight_shape] Incorrect shape != 2")
-        elif type(array[0]) is not int:
-            raise Exception("[validate_weight_shape] Dim 0 not integer")
-        elif type(array[1]) is not list:
-            raise Exception("[validate_weight_shape] Dim 1 not list")
-        elif type(array[1][0]) is not np.ndarray:
-            raise Exception("[validate_weight_shape] Dim 1:0 not ndarray")
-        else:
-            file.seek(0)
-            return True
+        Save model weights in compressed NPZ format
 
-    def file_upload(self, file_path):
+        Args:
+            file_path: Output path (.npz extension added automatically)
+            weighting: Weight for aggregation
+                - Number of training samples (most common)
+                - Fixed value for equal weighting (e.g., 1)
+                - Quality-based weight (e.g., validation_accuracy * num_samples)
+            parameters: List of numpy arrays containing model parameters
+
+        Returns:
+            str: Full file path with .npz extension
+
+        Example:
+            >>> file_path = client.save_weights(
+            ...     "model_weights",
+            ...     weighting=1000,
+            ...     parameters=[layer1, layer2, ...]
+            ... )
+            >>> print(file_path)  # "model_weights.npz"
+        """
+        if not file_path.endswith('.npz'):
+            file_path += '.npz'
+
+        # Create dict for savez_compressed
+        arrays_dict = {'weighting': np.array([weighting], dtype=np.float64)}
+        for i, arr in enumerate(parameters):
+            arrays_dict[f'layer_{i}'] = arr
+
+        np.savez_compressed(file_path, **arrays_dict)
+        return file_path
+
+    def validate_weight_shape(self, file) -> bool:
+        """
+        Validate NPZ file format before uploading to API
+
+        Args:
+            file: Binary file object (opened NPZ file)
+
+        Returns:
+            bool: True if validation passes
+
+        Raises:
+            Exception: If file format is invalid:
+                - Missing 'weighting' field
+                - Invalid weighting format (must be single value)
+                - No layer arrays found
+                - Layer arrays are not numpy arrays
+
+        Note:
+            File position is reset to beginning after validation
+        """
+        # Load NPZ without pickle
+        data = np.load(file, allow_pickle=False)
+
+        # Check for required 'weighting' field
+        if 'weighting' not in data.files:
+            raise Exception("[validate_weight_shape] Missing 'weighting' field in NPZ file")
+
+        weighting = data['weighting']
+        if weighting.size != 1:
+            raise Exception("[validate_weight_shape] 'weighting' must be a single value")
+
+        # Check that there are layer arrays
+        layer_keys = [k for k in data.files if k.startswith('layer_')]
+        if len(layer_keys) == 0:
+            raise Exception("[validate_weight_shape] No layer arrays found in NPZ file")
+
+        # Validate each layer is a numpy array
+        for key in layer_keys:
+            if not isinstance(data[key], np.ndarray):
+                raise Exception(f"[validate_weight_shape] {key} is not a numpy array")
+
+        file.seek(0)
+        return True
+
+    def file_upload(self, file_path: str) -> str:
+        """
+        Upload model weights file to BranchKey server
+
+        Args:
+            file_path: Path to NPZ file containing model weights
+
+        Returns:
+            str: Unique file ID assigned by server
+
+        Raises:
+            Exception: If upload fails or run is blocked
+
+        Note:
+            - File is validated before upload
+            - Waits for run to start if wait_for_run=True
+            - Requires run_status to be "start"
+
+        Example:
+            >>> file_path = client.save_weights("model", 1000, parameters)
+            >>> file_id = client.file_upload(file_path)
+            >>> print(f"Uploaded: {file_id}")
+        """
         try:
             if self.__run_status != "start":
                 if self.__wait_for_run:
@@ -197,7 +314,6 @@ class Client:
                     verify=self.__verify,
                     proxies=self.__proxies,
                 )
-                f.close()
 
                 if resp.status_code != HTTPStatus.CREATED:
                     raise Exception(resp.json()["error"])
@@ -210,7 +326,30 @@ class Client:
             )
             raise Exception(msg)
 
-    def file_download(self, file_id):
+    def file_download(self, file_id: str) -> bool:
+        """
+        Download aggregated model weights from server
+
+        Args:
+            file_id: Unique file ID (typically from aggregation notification)
+
+        Returns:
+            bool: True if download successful
+
+        Raises:
+            Exception: If download fails or file is empty
+
+        Note:
+            - File saved to ./aggregated_files/{file_id}.npz
+            - Directory created automatically if it doesn't exist
+            - File contains only layer arrays (no weighting field)
+
+        Example:
+            >>> if not client.queue.empty():
+            ...     aggregation_id = client.queue.get(block=False)
+            ...     client.file_download(aggregation_id)
+            ...     # File saved to ./aggregated_files/{aggregation_id}.npz
+        """
         try:
             if not os.path.exists(AGGREGATED_OUTPUT_DIR):
                 os.makedirs(AGGREGATED_OUTPUT_DIR)
@@ -239,7 +378,7 @@ class Client:
             if len(resp.content) == 0:
                 raise Exception("file not received")
 
-            f = open(AGGREGATED_OUTPUT_DIR + "/" + file_id + ".npy", "wb")
+            f = open(AGGREGATED_OUTPUT_DIR + "/" + file_id + ".npz", "wb")
             f.write(resp.content)
             f.close()
             print(f"downloaded file {file_id}")
@@ -251,7 +390,27 @@ class Client:
             )
             raise Exception(msg)
 
-    def send_performance_metrics(self, aggregation_id: str, data: str, mode: str):
+    def send_performance_metrics(self, aggregation_id: str, data: str, mode: str) -> bool:
+        """
+        Submit training or testing performance metrics to server
+
+        Args:
+            aggregation_id: ID of the aggregation these metrics relate to
+            data: JSON string containing metrics (e.g., '{"accuracy": 0.95, "loss": 0.12}')
+            mode: Metric type - "test", "train", or "non-federated"
+
+        Returns:
+            bool: True if submission successful, False otherwise
+
+        Example:
+            >>> import json
+            >>> metrics = {"accuracy": 0.95, "loss": 0.12}
+            >>> client.send_performance_metrics(
+            ...     aggregation_id="agg-uuid",
+            ...     data=json.dumps(metrics),
+            ...     mode="test"
+            ... )
+        """
         try:
             url = self.__api_host + "/api/v1/performance/analysis"
             headers = self._get_auth_headers()
@@ -274,7 +433,7 @@ class Client:
                 proxies=self.__proxies,
             )
 
-            print(f"received resp code form performance_analyser: {resp.status_code}")
+            print(f"received resp code from performance_analyser: {resp.status_code}")
             if resp.status_code != HTTPStatus.CREATED:
                 raise Exception(resp.json())
             else:
@@ -319,8 +478,23 @@ class Client:
     def run_number(self):
         return self.__run_number
 
-    def close(self):
-        """Gracefully shutdown the client and RabbitMQ consumer"""
+    def close(self) -> None:
+        """
+        Gracefully shutdown the client and RabbitMQ consumer
+
+        Stops the background RabbitMQ consumer thread and cleans up resources.
+        Should be called when done using the client, or use context manager.
+
+        Example:
+            >>> client = Client(credentials)
+            >>> # ... do work ...
+            >>> client.close()
+
+            Or with context manager:
+            >>> with Client(credentials) as client:
+            ...     # ... do work ...
+            ...     pass  # Automatically closed
+        """
         if self.consumer is not None and self.consumer.is_alive():
             try:
                 # Request consumer to stop using pika's thread-safe callback mechanism
